//////////////////////////////////////////////////////////////////////////
// I2C.h
// I2C
//////////////////////////////////////////////////////////////////////////

#include "i2c.h"
#include "TW68XX.h"

//===========================================================================
// I2C Class Implementation
//===========================================================================

/////////////////////////////////////////////////////////////////////////////
// Constructor
/////////////////////////////////////////////////////////////////////////////
CI2C::CI2C() :
_cycle(0L),
_errNum(I2CERR_OK),
_i2c_mode(I2CMode_None)
{
}

bool CI2C::I2CInitSWMode(long freq)
{
    // need to calibrate in order to generate the correct clock _cycle
    // the approach is to calculate how many dummy loop we need in order to
    // generate a _cycle that is 2 * freq * 1000 Hz
    _cycle = 10000L;    // use a large number to start
    ULONG elapsed = 0L;

    while(elapsed < 5) // loop until delay is long enough for calculation
    {
        _cycle *= 10;
        ULONG start = GetTickCount();

        for(volatile ULONG i = _cycle; i > 0; i--) ;
        elapsed = GetTickCount() - start;
    }

    if(freq > 1)
        _cycle = _cycle / elapsed * 1000L / freq / 2;

	// I2C is running software _i2c_mode
	_i2c_mode = I2CMode_SW;
	_errNum = I2CERR_OK;

	return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWStart( void )
//  Purpose: Generate START condition using software control
//  Input:   None
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWStart(void)
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }
	
	// SendStart - send an I2c start
	// i.e. SDA 1 -> 0 with SCL = 1

	if ( !I2CSWSetSDA( LevelHi )  ) { _errNum = I2CERR_SDA; return false; }
	if ( !I2CSWSetSCL( LevelHi )  ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelLow ) ) { _errNum = I2CERR_SDA; return false; }
	if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
	_errNum = I2CERR_OK;

    return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWStop( void )
//  Purpose: Generate STOP condition using software control
//  Input:   None
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWStop(void)
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

	// SendStop - sends an I2C stop, releasing the bus.
	// i.e. SDA 0 -> 1 with SCL = 1
	
	if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelLow ) )	{ _errNum = I2CERR_SDA; return false; }
	if ( !I2CSWSetSCL( LevelHi ) )	{ _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelHi )  )	{ _errNum = I2CERR_SDA; return false; }
   	_errNum = I2CERR_OK;
	
	return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWRead( UCHAR * value )
//  Purpose: Read a byte from the slave
//  Input:   None
//  Output:  UCHAR * value - byte read from slave
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWRead(UCHAR *value)
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

    *value = 0x00;

    // read 8 bits from I2c into Accumulator
    for(UCHAR mask = 0x80; mask > 0; mask = (UCHAR)(mask >> 1)) {
		if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
		if ( !I2CSWSetSCL( LevelHi ) )	{ _errNum = I2CERR_SCL; return false; }
		if ( I2CSWReadSDA() == LevelHi ){ *value=(UCHAR)( *value | mask ); }// set the bit
    }
	_errNum = I2CERR_OK;
	
    return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWWrite( UCHAR value )
//  Purpose: Write a byte to the slave
//  Input:   UCHAR value - byte to be written to slave
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWWrite(UCHAR value)
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

   // generate bit patterns by setting SCL and SDA lines
    for(UCHAR mask = 0x80; mask > 0; mask = (UCHAR)(mask >> 1)) {
        if(!I2CSWSetSCL(LevelLow)) {_errNum = I2CERR_SCL; return false;}

        // Send one data bit.
        if(value & mask) { // Put data bit on pin.
            if(!I2CSWSetSDA(LevelHi)) {_errNum = I2CERR_SDA; return false;}
        }
        else if(!I2CSWSetSDA(LevelLow)) {_errNum = I2CERR_SDA; return false;}

        if(!I2CSWSetSCL(LevelHi)) {_errNum = I2CERR_SCL; return false;}
    }
	
    return I2CSWWaitForACK();
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWSendACK( void )
//  Purpose: Send ACK to slave
//  Input:   None
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWSendACK( void )
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

	// Generate ACK signal
	// i.e. SDA = 0 with SCL = 1

	if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelLow ) ) { _errNum = I2CERR_SDA; return false; }
	if ( !I2CSWSetSCL( LevelHi  ) ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelHi  ) ) { _errNum = I2CERR_SDA; return false; }
	_errNum = I2CERR_OK;
	
	return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWSendNACK( void )
//  Purpose: Send NACK to slave
//  Input:   None
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWSendNACK( void )
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

	// Generate NACK signal
	// i.e. SDA = 1 with SCL = 1

	if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelHi )  ) { _errNum = I2CERR_SDA; return false; }
	if ( !I2CSWSetSCL( LevelHi )  ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelLow ) ) { _errNum = I2CERR_SDA; return false; }
	_errNum = I2CERR_OK;
	
	return true;
}

bool CI2C::I2CSWSetSCL(TLevel scl)
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

    // loop until SCL really changes or timeout
    ULONG maxWait = 500;                  // 500 mSec
    ULONG startTime = GetTickCount();

    while(1)
    {
        // has SCL changed yet?
        if(I2CSWReadSCL() == scl)
            break;

        // timeout?
        if(GetTickCount() - startTime > (ULONG)maxWait)
        {
            _errNum = I2CERR_TIMEOUT;
            return false;
        }
    }

	I2CSWBitDelay();
	_errNum = I2CERR_OK;

    return true;
}

bool CI2C::I2CSWSetSDA(TLevel sda)
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }
	
    I2CSWBitDelay();
	_errNum = I2CERR_OK;

	UNREFERENCED_PARAMETER(sda);
	
    return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  void CI2C::I2CSWBitDelay( void )
//  Purpose: Insures minimum high and low clock times on I2C bus.
//  Input:   None
//  Output:  None
//  Return:  None
//  Note:    This routine must be tuned for the desired frequency.
/////////////////////////////////////////////////////////////////////////////
void CI2C::I2CSWBitDelay( void )
{
	volatile ULONG i;
	for ( i = _cycle; i > 0; i-- ) ;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWWaitForACK( void )
//  Purpose: Determine if ACK is received in software _i2c_mode
//  Input:   None
//  Output:  None
//  Return:  true if ACK received; false if timeout
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWWaitForACK( void )
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

	if ( !I2CSWSetSCL( LevelLow ) ){ _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelHi ) ) { _errNum = I2CERR_SDA; return false; }

	// loop until either ACK or timeout
	ULONG maxWait = 500; // 500 mSec
	ULONG startTime = GetTickCount();

	while (1)
	{
		// SDA pin == 0 means the slave ACK'd
		if ( I2CSWReadSDA() == 0 )
		{
			if ( !I2CSWSetSCL( LevelHi ) ) { _errNum = I2CERR_SCL; return false; }
			if ( !I2CSWSetSCL( LevelLow ) ){ _errNum = I2CERR_SCL; return false; }
			_errNum = I2CERR_OK;
			return true;
		}

		// timeout?
		if ( GetTickCount() - startTime > (ULONG)maxWait )
		{
			I2CSWStop();
			_errNum = I2CERR_TIMEOUT;
			return false;
		}
	} // while

	return true;
}

//--------------------------------------------------------
// ʱ
//--------------------------------------------------------
ULONG CI2C::GetTickCount( void )
{
    ULONGLONG ticks;
    ULONGLONG rate;
	
    ticks = (ULONGLONG)KeQueryPerformanceCounter((PLARGE_INTEGER)&rate).QuadPart;
	
    //
    // convert from ticks to 100ns clock
    //
    ticks = (ticks & 0xFFFFFFFF00000000) / rate * 10000000 + (ticks &
		0xFFFFFFFF) * 10000000 / rate;

	return (ULONG)( ticks / 10000 );
}

//--------------------------------------------------------
// д
//--------------------------------------------------------
bool CI2C::i2cWriteDataNoLock(int nLength, UCHAR *pData)
{
	int   i;
	LONG  err=0;

	if(I2CSWStart()) {
 		for(i = 0; i < nLength;i++)	{
			if(!I2CSWWrite(*(pData+i)))	{
				break;
			}
		}
	}

	err = I2CGetLastError();
	I2CSWStop();
	//Sleep(10);
	
	return (err==I2CERR_OK);
}

//--------------------------------------------------------
// д
//--------------------------------------------------------
bool CI2C::i2cWriteDataNoStopNoLock(int nLength, UCHAR *pData)
{
	int   i;
	LONG  err;

	if(I2CSWStart()) {
		for(i = 0; i < nLength;i++)	{
			if(!I2CSWWrite(*(pData+i)))	{
				break;
			}
		}
	}
	
	err = I2CGetLastError();
	
	return (err==I2CERR_OK);
}

//--------------------------------------------------------
// 
//--------------------------------------------------------
bool CI2C::i2cReadDataNoLock(int nWriteBytes, UCHAR *cWrite, int nReadBytes, UCHAR *pData)
{	
	LONG  err=0;
	int   i;
	
	if(I2CSWStart()) {
		for(i = 0; i < nWriteBytes;i++) {
			if(!I2CSWWrite(cWrite[i]))	{
				break;
			}
		}
	
		if(i >= nWriteBytes) {
			for(i = 0; i < nReadBytes; i++) {
				if(!I2CSWRead(pData+i)) {
					break;
				}
				if(i < nReadBytes - 1) {
					I2CSWSendACK();
				}
			}
			I2CSWSendNACK();
		}
	}

	err = I2CGetLastError();
	I2CSWStop();
	//Sleep(10);
	
	return (err==I2CERR_OK);
}

//--------------------------------------------------------
// 
//--------------------------------------------------------
bool CI2C::i2cReadDataNoLock(UCHAR *cWrite, int nWirteBytes, UCHAR* pData, int nReadBytes)
{
	bool ret = i2cWriteDataNoLock(nWirteBytes, cWrite);
	if(ret) {
		cWrite[0] |= 1;
		ret = i2cReadDataNoLock(1, cWrite, nReadBytes, pData);
	}

	return ret;
}

//--------------------------------------------------------
// д
//--------------------------------------------------------
bool CI2C::i2cWriteByteNoLock(UCHAR addr, UCHAR sub, UCHAR mask, UCHAR val)
{
	UCHAR cWrite[3] = {addr, sub, 0};
	bool  ret = i2cReadDataNoLock(cWrite, 2, &cWrite[2], 1);

	if(ret) {
		cWrite[0]  = addr;
		cWrite[2] &= ~mask;
		cWrite[2] |= (val & mask);
		ret = i2cWriteDataNoLock(3, cWrite);
	}

	return ret;
}

//--------------------------------------------------------
// д
//--------------------------------------------------------
bool CI2C::i2cWriteData(int nLength, UCHAR *pData)
{
	CScopedLock autoLock(&_lock, 0);

	if(!autoLock.Status()) {
		_errNum = I2CERR_TIMEOUT;
		return false;
	}

	return i2cWriteDataNoLock(nLength, pData);
}

//--------------------------------------------------------
// д
//--------------------------------------------------------
bool CI2C::i2cWriteDataNoStop(int nLength, UCHAR *pData)
{
	CScopedLock autoLock(&_lock, 0);
	
	if(!autoLock.Status()) {
		_errNum = I2CERR_TIMEOUT;
		return false;
	}
	
	return i2cWriteDataNoStopNoLock(nLength, pData);
}

//--------------------------------------------------------
// 
//--------------------------------------------------------
bool CI2C::i2cReadData(int nWriteBytes, UCHAR *cWrite, int nReadBytes, UCHAR *pData)
{	
	CScopedLock autoLock(&_lock, 0);
	
	if(!autoLock.Status()) {
		_errNum = I2CERR_TIMEOUT;
		return false;
	}
	
	return i2cReadDataNoLock(nWriteBytes, cWrite, nReadBytes, pData);
}

//--------------------------------------------------------
// 
//--------------------------------------------------------
bool CI2C::i2cReadData(UCHAR *cWrite, int nWirteBytes, UCHAR* pData, int nReadBytes)
{
	CScopedLock autoLock(&_lock, 0);
	
	if(!autoLock.Status()) {
		_errNum = I2CERR_TIMEOUT;
		return false;
	}

	return i2cReadDataNoLock(cWrite, nWirteBytes, pData, nReadBytes);
}

//--------------------------------------------------------
// д
//--------------------------------------------------------
bool CI2C::i2cWriteByte(UCHAR addr, UCHAR sub, UCHAR mask, UCHAR val)
{
	CScopedLock autoLock(&_lock, 0);
	
	if(!autoLock.Status()) {
		_errNum = I2CERR_TIMEOUT;
		return false;
	}

	return i2cWriteByteNoLock(addr, sub, mask, val);
}

//--------------------------------------------------------
// Ժ
//--------------------------------------------------------
bool CI2C::i2cTest(UCHAR I2CAdd)
{	
	if(!I2CSWStart())		return false;
	if(!I2CSWWrite(I2CAdd)) return false;
	if(!I2CSWStop())		return false;

	return true;	
}

//--------------------------------------------------------
// I2Cд24C02Ժ
//--------------------------------------------------------
bool CI2C::i2cTest24C02()
{
	int i;
	UCHAR data[5] = {0x60, 0x08, 0xB0, 0x01};
	
	//
	// д
	//

	if(I2CSWStart()) { 

		//д24C02ַ
		if(I2CSWWrite(0xA0)) { 

			//дsub addr
			if(I2CSWWrite(0xFC)) { 

				//д
				for(i=0; i<4; i++) {
					if(!I2CSWWrite(data[i])) { break; }
				}
			}
		}
	}

	if(!I2CSWStop()) { return false; }
	//Sleep(40);

	//
	// 
	//

	if(I2CSWStart()) {

		//д24C02ַ
		if(I2CSWWrite(0xA0)) {

			//дsub addr
			if(I2CSWWrite(0xFC)) {

				if(I2CSWStart()) {
					//дaddr
					if(I2CSWWrite(0xA1)) {

						//
						for(i=0; i<4; i++) {
							if(!I2CSWRead(data+i)) { 
								break;
							}
							else {
								I2CSWSendACK();
							}
						}
						I2CSWSendNACK();
					}
				}
			}
		}
	}

	if(!I2CSWStop()) { return false; }

	return true;
}