// I2C6864.cpp: implementation of the CI2C6864 class.
//
//////////////////////////////////////////////////////////////////////

#include "I2C6864.h"
#include "TW686X.h"
#include "TW68XX.h"

#define TW6864_I2C_SCLK_BIT			BIT(0)
#define TW6864_I2C_SDA_DATA_BIT		BIT(1)
#define TW6864_I2C_SCLK_CTRL_BIT	BIT(16)  //1:Input,0=Output
#define TW6864_I2C_SDA_CTRL_BIT		BIT(17)  //1:Input,0=Output

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

bool CI2C6864::i2cInit(PVOID pMem, HANDLE hLock)
{ 
	m_pMem=(KMemoryRange*)pMem; 
	if(m_pMem == NULL) { return false; }

	control_word = m_pMem->ind(TW6864_R_GPIO_REG);
	_lock.Initialize(hLock);

	return true; 
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C6864::I2CSWSetSCL( Level scl )
//  Purpose: Set software SCL value
//  Input:   Level scl - Hi releases the SCL output; Low forces the SCL output low
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C6864::I2CSWSetSCL( TLevel scl )
{
	control_word = m_pMem->ind(TW6864_R_GPIO_REG);
	control_word &= ~TW6864_I2C_SCLK_CTRL_BIT; 
	if(scl == LevelHi) {
		control_word |= TW6864_I2C_SCLK_BIT;
	}
	else {
		control_word &= ~TW6864_I2C_SCLK_BIT;
	}

	m_pMem->outd(TW6864_R_GPIO_REG, control_word);

	return CI2C::I2CSWSetSCL(scl);
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C6864::I2CSWSetSDA( Level sda )
//  Purpose: Set software SDA value
//  Input:   Level sda - Hi releases the SDA output; Low forces the SDA output low
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C6864::I2CSWSetSDA( TLevel sda )
{
	control_word = m_pMem->ind(TW6864_R_GPIO_REG);
	control_word &= ~TW6864_I2C_SDA_CTRL_BIT; 
	if(sda == LevelHi) {
		control_word |= TW6864_I2C_SDA_DATA_BIT;	
	}
	else {
		control_word &= ~TW6864_I2C_SDA_DATA_BIT;
	}

	m_pMem->outd(TW6864_R_GPIO_REG, control_word);

	return CI2C::I2CSWSetSDA(sda);
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  int CI2C6864::I2CSWReadSCL( void )
//  Purpose: Read software SCL value
//  Input:   None
//  Output:  None
//  Return:  SCL value
/////////////////////////////////////////////////////////////////////////////
int CI2C6864::I2CSWReadSCL( void )
{
	//control_word |= TW6864_I2C_SCLK_CTRL_BIT; 
	//m_pMem->outd(TW6864_R_GPIO_REG, control_word);

	return (m_pMem->ind(TW6864_R_GPIO_REG) & TW6864_I2C_SCLK_BIT);
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  int CI2C6864::I2CSWReadSDA( void )
//  Purpose: Read software SDA value
//  Input:   None
//  Output:  None
//  Return:  SDA value
/////////////////////////////////////////////////////////////////////////////
int CI2C6864::I2CSWReadSDA( void )
{
	control_word = m_pMem->ind(TW6864_R_GPIO_REG);
	control_word |= TW6864_I2C_SDA_CTRL_BIT; 
	m_pMem->outd(TW6864_R_GPIO_REG, control_word);

	return ((m_pMem->ind(TW6864_R_GPIO_REG) & TW6864_I2C_SDA_DATA_BIT)>>1);
}

//--------------------------------------------------------
// 
//--------------------------------------------------------
bool CI2C6864::i2cReadDataNoLock(UCHAR *cWrite, int nWirteBytes, UCHAR* pData, int nReadBytes)
{
	bool  ret = false;
	UCHAR data= 0;

	if((cWrite[0]==TW2864_R_I2C_ADDR0) && (cWrite[1]==TW2864_R_VERTICAL_SHARPNESS)) {
		if(m_shadow_0x85 != 0xFF) {
			data = m_shadow_0x85;
			ret = true;
		}
	}

	if(!ret) {
		if(i2cWriteDataNoLock(nWirteBytes, cWrite)) {
			cWrite[0] |= 1;
			ret = CI2C::i2cReadDataNoLock(1, cWrite, nReadBytes, &data);
		}
	}

	if(ret) {
		*pData = data;
	}
	
	return ret;
}

//--------------------------------------------------------
// д
//--------------------------------------------------------
bool CI2C6864::i2cWriteByteNoLock(UCHAR addr, UCHAR sub, UCHAR mask, UCHAR val)
{
	UCHAR cWrite[3] = {addr, sub, 0};
	bool  ret = true;
	
	if(mask != 0xFF) {
		ret = i2cReadDataNoLock(cWrite, 2, &cWrite[2], 1);
	}
	
	if(ret) {
		cWrite[0]  = addr;
		cWrite[2] &= ~mask;
		cWrite[2] |= (val & mask);
		ret = i2cWriteDataNoLock(3, cWrite);
		if(ret && (sub==TW2864_R_VERTICAL_SHARPNESS) && (addr==TW2864_R_I2C_ADDR0)) {
			m_shadow_0x85 = cWrite[2];
		}
	}
	
	return ret;
}

//--------------------------------------------------------
// 
//--------------------------------------------------------
bool CI2C6864::i2cReadData(UCHAR *cWrite, int nWirteBytes, UCHAR* pData, int nReadBytes)
{
	bool  ret = false;
	UCHAR data= 0;
	CScopedLock autoLock(&_lock, 0);
	
	if(!autoLock.Status()) {
		_errNum = I2CERR_TIMEOUT;
		return false;
	}
	
	if((cWrite[0]==TW2864_R_I2C_ADDR0) && (cWrite[1]==TW2864_R_VERTICAL_SHARPNESS)) {
		if(m_shadow_0x85 != 0xFF) {
			data = m_shadow_0x85;
			ret = true;
		}
	}
	
	if(!ret) {
		if(i2cWriteDataNoLock(nWirteBytes, cWrite)) {
			cWrite[0] |= 1;
			ret = CI2C::i2cReadDataNoLock(1, cWrite, nReadBytes, &data);
		}
	}
	
	if(ret) {
		*pData = data;
	}
	
	return ret;
}

//--------------------------------------------------------
// д
//--------------------------------------------------------
bool CI2C6864::i2cWriteByte(UCHAR addr, UCHAR sub, UCHAR mask, UCHAR val)
{
	UCHAR cWrite[3] = {addr, sub, 0};
	bool  ret = true;
	CScopedLock autoLock(&_lock, 0);
	
	if(!autoLock.Status()) {
		_errNum = I2CERR_TIMEOUT;
		return false;
	}
	
	if(mask != 0xFF) {
		ret = i2cReadDataNoLock(cWrite, 2, &cWrite[2], 1);
	}
	
	if(ret) {
		cWrite[0]  = addr;
		cWrite[2] &= ~mask;
		cWrite[2] |= (val & mask);
		ret = i2cWriteDataNoLock(3, cWrite);
		if(ret && (sub==TW2864_R_VERTICAL_SHARPNESS) && (addr==TW2864_R_I2C_ADDR0)) {
			m_shadow_0x85 = cWrite[2];
		}
	}
	
	return ret;
}