// kdmqueue.cpp - implementation of class KDriverManagedQueue
//=============================================================================
//
// Compuware Corporation
// NuMega Lab
// 9 Townsend West
// Nashua, NH 03060  USA
//
// Copyright (c) 1998 Compuware Corporation. All Rights Reserved.
// Unpublished - rights reserved under the Copyright laws of the
// United States.
//
// U.S. GOVERNMENT RIGHTS-Use, duplication, or disclosure by the
// U.S. Government is subject to restrictions as set forth in 
// Compuware Corporation license agreement and as provided in 
// DFARS 227.7202-1(a) and 227.7202-3(a) (1995), 
// DFARS 252.227-7013(c)(1)(ii)(OCT 1988), FAR 12.212(a) (1995), 
// FAR 52.227-19, or FAR 52.227-14 (ALT III), as applicable.  
// Compuware Corporation.
// 
// This product contains confidential information and trade secrets 
// of Compuware Corporation. Use, disclosure, or reproduction is 
// prohibited without the prior express written permission of Compuware 
// Corporation.
//
//=============================================================================

#include <vdw.h>

///////////////////////////////////////////////////////////////////////////////
//
//	KDriverManagedQueue(void)
//
//	Routine Description:
//
//		Constructor for class KDriverManagedQueueEx
//
//	Parameters:
//
//		None
//
//	Return Value:
//
//		None
//
//	Comments:
//
//		This is a subclass of KDeviceQueue. It is also an abstract class,
//		due to member StartIo.
//
KDriverManagedQueue::KDriverManagedQueue(void) :
	KDeviceQueue(),
	m_CurrentIrp(0)
{
}

///////////////////////////////////////////////////////////////////////////////
//
//	NTSTATUS
//	QueueIrp(
//		KIrp I )
//
//	Routine Description:
//
//		Queues an IRP to the StartIo routine. If the queue is empty, calls
//		StartIo synchronously.
//
//	Parameters:
//
//		I - The IRP to queue (or process).
//
//	Return Value:
//
//		Always returns STATUS_PENDING.
//
//	Comments:
//
//		If the queue is not busy, this member function raises IRQL to
//		DISPATCH_LEVEL and calls memeber function StartIo. If the queue is
//		busy, the IRP is put on the tail of the queue. Queuing an IRP when
//		the queue is not busy causes the queue to become busy. It remains
//		busy until there is an attempt to dequeue an IRP from the queue
//		when it is empty. Member function NextIrp dequeues IRPs. IRPs
//		queued by this member function are cancelable. QueueIrp sets the
//		cancel routine in the IRP such that overridable virtual member
//		function Cancel is called if a higher level driver cancels the
//		IRP.
//
//		The system calls the cancel routine with only the device object and
//		the IRP, so the driver has no way of knowing which of its private
//		queues the IRP is in. To overcome this, this routine stores the
//		address of the KDriverManagedQueue object in the Information field
//		of the IRP, where it can be retrieved by the cancel routine.
//
NTSTATUS KDriverManagedQueue::QueueIrp(KIrp I)
{
	CancelSpinLock::Acquire();
	I.Information() = (ULONG_PTR)this;
	I.SetCancelRoutine(CancelDispatch);
	CancelSpinLock::Release();
	I.MarkPending();

	KIRQL irql;
	
	KeRaiseIrql(DISPATCH_LEVEL, &irql);		

	if ( !Insert(I) )
	{
		m_CurrentIrp = I;
		StartIo(I);
	}

	KeLowerIrql(irql);

	return STATUS_PENDING;
}

///////////////////////////////////////////////////////////////////////////////
//
//	VOID
//	NextIrp(
//		KIrp I,
//		CCHAR boost )
//
//	Routine Description:
//
//		Complete the specified IRP and start processing the next IRP, if present,
//		in the queue.
//
//	Parameters:
//
//		I - IRP to be completed.
//
//		boost - The increment in priority to be granted to the thread waiting for
//			the IRP to complete. The default is IO_NO_INCREMENT.
//
//	Return Value:
//
//		None
//
//	Comments:
//
//		For WDM drivers that track outstanding requests, use PnpNextIrp rather
//		than this function.
//
//		When a driver is ready to complete the current IRP and dequeue the
//		next, this member is called to perform both those operations. The
//		caller must set the Status and Information fields of the IRP being
//		completed before calling this function. A typical fragment of code
//		looks like:
//													
//			I.Status() = STATUS_SUCCESS;
//			I.Information() = nBytesReturned;
//			NextIrp(I);
//
//		To only start processing the next IRP, without completing any IRP,
//		pass the first parameter as KIrp(0). NextIrp attempts to dequeue an
//		IRP from the embedded device queue (remember, KDriverManagedQueue
//		is a subclass of KDeviceQueue). If it succeeds, data member
//		m_CurrentIrp is set to the dequeued IRP,  IRQL is raised to
//		DISPATCH_LEVEL, and pure virtual member function StartIo is called.
//		Otherwise, m_CurrentIrp is set to NULL.
//
VOID KDriverManagedQueue::NextIrp(KIrp I, CCHAR boost)
{
	KIRQL irql;

	// Acquire the global cancel spinlock.
	CancelSpinLock::Acquire();

	// Clear the cancel routine, since we will be completing this IRP.
	// The user may have already cleared the cancel routine, but it doesn't
	// hurt to do it again.
	if ( !I.IsNull() )
	{
		I.SetCancelRoutine(NULL);
	}		

	// Dequeue the next IRP. Removal when queue not busy is an error
	if (m_pdq->Busy)
	{
		m_CurrentIrp = RemoveHead();
		CancelSpinLock::Release();

		if ( !m_CurrentIrp.IsNull() )
		{
			KeRaiseIrql(DISPATCH_LEVEL, &irql);
			StartIo(m_CurrentIrp);
			KeLowerIrql(irql);
		}
	}
	else
	{
	// This should never happen.
		ASSERT(FALSE);
		m_CurrentIrp.m_Irp = 0;
		CancelSpinLock::Release();
	}

	// Complete the supplied IRP	
	if ( !I.IsNull() )
		I.Complete(I.Status(), boost);
}

#if _WDM_
VOID KDriverManagedQueue::PnpNextIrp(KPnpDevice* pDevice, KIrp I, CCHAR boost)
{
	NextIrp(I, boost);		 
	if ( !I.IsNull() )
		pDevice->DecrementOutstandingRequestCount();
}
#endif // _WDM_

///////////////////////////////////////////////////////////////////////////////
//
//	VOID
//	Cancel(
//		KIrp I)
//
//	Routine Description:
//
//		Overridable function called by the system to cancel a request.
//
//	Parameters:
//
//		I - The IRP to be canceled.
//
//	Return Value:
//
//		None
//
//	Comments:
//
//		The class always sets up this routine as the cancel routine. When
//		called, the system wide cancel spin lock is held. If the driver has
//		special cancellation requirements, it can override this function.
//		Note also that this cancel routine is intended to be used primarily
//		while the IRP is queued, waiting to be processed.  Once processing begins,
//		usually the IRP is taken out of this cancelable state (see
//		KIrp::TestAndSetCancelRoutine()).  The driver writer may use other cancel
//		routines in the course of processing the IRP.
//
//		Important: A WDM driver needs to needs to take into account the
//		PnP aspects of IRP management, such as IRP reference counting.
//		Use PnpNextIrp instead NextIrp of, and KIrp::PnpComplete instead of
//		KIrp::Complete. Refer to the default implementation as an example
//		if overriding this function.
//
VOID KDriverManagedQueue::Cancel(KIrp I)
{

#if _WDM_
	// Cast the device object pointer passed in to a KDevice pointer
	// to find out if the device object is a Pnp device or not.  If it
	// is a Pnp device, we will call the appropriate Pnp functions below
	// to maintain proper IRP reference counting.
	PDEVICE_OBJECT pDev = I.DeviceObject();
	ASSERT(pDev);
	KDevice * pDevice = KDevicePTR(pDev);
	KPnpDevice * pPnpDevice = NULL;
	if (pDevice->IsPnpDevice())
	{
		pPnpDevice = static_cast <KPnpDevice *> (pDevice);
	}
#endif // _WDM_

	// Test if the IRP is the current IRP.
	if ( (PIRP)I == CurrentIrp() )
	{

	// If so, NULL it out, release the global cancel spinlock, then
	// complete it as canceled and start the next IRP.  Note that
	// if we got here, the IRP was still in a cancelable state when
	// it was canceled, and most likely is just about to start being
	// processed on the device.
		CurrentIrp() = NULL;
		CancelSpinLock::Release(I.CancelIrql());
		I.Information() = 0;
		I.Status() = STATUS_CANCELLED;

#if _WDM_
		if (pPnpDevice == NULL)
		{
			NextIrp(I, IO_NO_INCREMENT);
		}
		else
		{
			PnpNextIrp(pPnpDevice, I, IO_NO_INCREMENT);
		}
#else // _WDM_
		NextIrp(I, IO_NO_INCREMENT);
#endif // _WDM_

	}
	// See if the IRP can be removed from the device queue.
	else if (RemoveSpecificEntry(I))
	{

	// If so, release the global cancel spinlock and complete it
	// as canceled.
		CancelSpinLock::Release(I.CancelIrql());
		I.Information() = 0;

#if _WDM_
		if (pPnpDevice == NULL)
		{
			I.Complete(STATUS_CANCELLED);
		}
		else
		{
			I.PnpComplete(pPnpDevice, STATUS_CANCELLED);
		}
#else // _WDM_
		I.Complete(STATUS_CANCELLED);
#endif // _WDM_

	}
	else
	{

	// If we got here the IRP wasn't the current IRP, and wasn't in the
	// device queue.  This could happen if the IRP was removed by the
	// device queue's cleanup routine, just as it was being canceled.
	// In this case we have no work to do, since the cleanup routine
	// will cancel the IRP.  Simply release the global cancel spinlock.
		CancelSpinLock::Release(I.CancelIrql());
	}
}

///////////////////////////////////////////////////////////////////////////////
//
//	VOID
//	CancelDispatch(
//		PDEVICE_OBJECT pDev,
//		PIRP pIrp )
//
//	Routine Description:
//
//		Static to dispatch to virtual member function.
//
//	Parameters:
//
//		pDev - pointer to device object
//		pIrp - IRP to be canceled.
//
//	Return Value:
//
//		None
//
//	Comments:
//
//		QueueIrp stores the queue address in the Information field of
//		the IRP.
//
VOID KDriverManagedQueue::CancelDispatch(PDEVICE_OBJECT pDev, PIRP pIrp)
{
	KIrp I(pIrp);
	KDriverManagedQueue* pQ = (KDriverManagedQueue*)I.Information();

	pQ->Cancel(I);
	UNREFERENCED_PARAMETER(pDev);
}

