//////////////////////////////////////////////////////////////////////////
// I2C.h
// I2C
//////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "i2c.h"

ULONGLONG CI2C::_tk_freq = 0;

//===========================================================================
// I2C Class Implementation
//===========================================================================

/////////////////////////////////////////////////////////////////////////////
// Constructor
/////////////////////////////////////////////////////////////////////////////
CI2C::CI2C() :
_cycle(0L),
_errNum(I2CERR_OK),
_i2c_mode(I2CMode_None)
{
	if(_tk_freq == 0) {
		LARGE_INTEGER freq;
		if(QueryPerformanceFrequency(&freq)) {
			_tk_freq = (ULONGLONG)freq.QuadPart;
		}
		else {
			_trace(TEXT("[Ϣ]CI2C::QueryPerformanceFrequencyʧ\n"));
		}
	}
}

bool CI2C::I2CInitSWMode(long freq)
{
    // need to calibrate in order to generate the correct clock _cycle
    // the approach is to calculate how many dummy loop we need in order to
    // generate a _cycle that is 2 * freq * 1000 Hz
    _cycle = 10000L;    // use a large number to start
    ULONG elapsed = 0L;

    while(elapsed < 5) // loop until delay is long enough for calculation
    {
        _cycle *= 10;
        ULONG start = GetTickCount();

        for(volatile ULONG i = _cycle; i > 0; i--) ;
        elapsed = GetTickCount() - start;
    }

    if(freq > 1)
        _cycle = _cycle / elapsed * 1000L / freq / 2;

	// I2C is running software _i2c_mode
	_i2c_mode = I2CMode_SW;
	_errNum = I2CERR_OK;

	return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWStart( void )
//  Purpose: Generate START condition using software control
//  Input:   None
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWStart(void)
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }
	
	// SendStart - send an I2c start
	// i.e. SDA 1 -> 0 with SCL = 1

	if ( !I2CSWSetSDA( LevelHi )  ) { _errNum = I2CERR_SDA; return false; }
	if ( !I2CSWSetSCL( LevelHi )  ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelLow ) ) { _errNum = I2CERR_SDA; return false; }
	if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
	_errNum = I2CERR_OK;
    return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWStop( void )
//  Purpose: Generate STOP condition using software control
//  Input:   None
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWStop(void)
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

	// SendStop - sends an I2C stop, releasing the bus.
	// i.e. SDA 0 -> 1 with SCL = 1

	if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelLow ) )	{ _errNum = I2CERR_SDA; return false; }
	if ( !I2CSWSetSCL( LevelHi ) )	{ _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelHi )  )	{ _errNum = I2CERR_SDA; return false; }
   	_errNum = I2CERR_OK;
	return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWRead( BYTE * value )
//  Purpose: Read a byte from the slave
//  Input:   None
//  Output:  BYTE * value - byte read from slave
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWRead(UCHAR *value)
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

    *value = 0x00;

    // read 8 bits from I2c into Accumulator
    for(UCHAR mask = 0x80; mask > 0; mask = (UCHAR)(mask >> 1)) {
		if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
		if ( I2CSWReadSDA() == LevelHi ){ *value=(BYTE)( *value | mask ); }// set the bit
		if ( !I2CSWSetSCL( LevelHi ) )	{ _errNum = I2CERR_SCL; return false; }
    }
	_errNum = I2CERR_OK;

    return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWWrite( BYTE value )
//  Purpose: Write a byte to the slave
//  Input:   BYTE value - byte to be written to slave
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWWrite(UCHAR value)
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

    // generate bit patterns by setting SCL and SDA lines
    for(UCHAR mask = 0x80; mask > 0; mask = (UCHAR)(mask >> 1)) {
        if(!I2CSWSetSCL(LevelLow)) {_errNum = I2CERR_SCL; return false;}

        // Send one data bit.
        if(value & mask) { // Put data bit on pin.
            if(!I2CSWSetSDA(LevelHi)) {_errNum = I2CERR_SDA; return false;}
        }
        else if(!I2CSWSetSDA(LevelLow)) {_errNum = I2CERR_SDA; return false;}

        if(!I2CSWSetSCL(LevelHi)) {_errNum = I2CERR_SCL; return false;}
    }

    return I2CSWWaitForACK();
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWSendACK( void )
//  Purpose: Send ACK to slave
//  Input:   None
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWSendACK( void )
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

	// Generate ACK signal
	// i.e. SDA = 0 with SCL = 1

	if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelLow ) ) { _errNum = I2CERR_SDA; return false; }
	if ( !I2CSWSetSCL( LevelHi  ) ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelHi  ) ) { _errNum = I2CERR_SDA; return false; }
	_errNum = I2CERR_OK;
	return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWSendNACK( void )
//  Purpose: Send NACK to slave
//  Input:   None
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWSendNACK( void )
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

	// Generate NACK signal
	// i.e. SDA = 1 with SCL = 1

	if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelHi )  ) { _errNum = I2CERR_SDA; return false; }
	if ( !I2CSWSetSCL( LevelHi )  ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSCL( LevelLow ) ) { _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelLow ) ) { _errNum = I2CERR_SDA; return false; }
	_errNum = I2CERR_OK;
	return true;
}

bool CI2C::I2CSWSetSCL(TLevel scl)
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

    // loop until SCL really changes or timeout
    ULONG maxWait = 500;                  // 500 mSec
    ULONG startTime = TickCount();

    while(1)
    {
        // has SCL changed yet?
        if(I2CSWReadSCL() == scl)
            break;

        // timeout?
        if(TickCount() - startTime > (ULONG)maxWait)
        {
            _errNum = I2CERR_TIMEOUT;
            return false;
        }
    }

	I2CSWBitDelay();
	_errNum = I2CERR_OK;

    return true;
}

bool CI2C::I2CSWSetSDA(TLevel sda)
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }
	
    I2CSWBitDelay();
	_errNum = I2CERR_OK;
	
    return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  void CI2C::I2CSWBitDelay( void )
//  Purpose: Insures minimum high and low clock times on I2C bus.
//  Input:   None
//  Output:  None
//  Return:  None
//  Note:    This routine must be tuned for the desired frequency.
/////////////////////////////////////////////////////////////////////////////
void CI2C::I2CSWBitDelay( void )
{
	volatile ULONG i;
	for ( i = _cycle; i > 0; i-- ) ;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C::I2CSWWaitForACK( void )
//  Purpose: Determine if ACK is received in software _i2c_mode
//  Input:   None
//  Output:  None
//  Return:  true if ACK received; false if timeout
/////////////////////////////////////////////////////////////////////////////
bool CI2C::I2CSWWaitForACK( void )
{
	// check if correct _i2c_mode is selected
	if ( _i2c_mode != I2CMode_SW ) { _errNum = I2CERR_MODE; return false; }

	if ( !I2CSWSetSCL( LevelLow ) ){ _errNum = I2CERR_SCL; return false; }
	if ( !I2CSWSetSDA( LevelHi ) ) { _errNum = I2CERR_SDA; return false; }

	// loop until either ACK or timeout
	DWORD maxWait = 500; // 500 mSec
	DWORD startTime = TickCount();

	while (1)
	{
		// SDA pin == 0 means the slave ACK'd
		if ( I2CSWReadSDA() == 0 )
		{
			if ( !I2CSWSetSCL( LevelHi ) ) { _errNum = I2CERR_SCL; return false; }
			if ( !I2CSWSetSCL( LevelLow ) ){ _errNum = I2CERR_SCL; return false; }
			_errNum = I2CERR_OK;
			return true;
		}

		// timeout?
		if ( TickCount() - startTime > (DWORD)maxWait )
		{
			I2CSWStop();
			_errNum = I2CERR_TIMEOUT;
			return false;
		}
	} // while

	return true;
}

//--------------------------------------------------------
// ʱ
//--------------------------------------------------------
ULONG CI2C::TickCount( void )
{
	LARGE_INTEGER tk;
	ULONGLONG ticks;

	if(QueryPerformanceCounter(&tk)) {
		ticks = (ULONGLONG)tk.QuadPart;

		//
		// convert from ticks to 100ns clock
		//
		ticks = (ticks & 0xFFFFFFFF00000000) / _tk_freq * 10000000 +
				(ticks & 0xFFFFFFFF) * 10000000 / _tk_freq;

		return (ULONG)( ticks / 10000 );
	}

	return 0;
}

//========================================================================
void CI2C::WaitTick( ULONG tick_cnt)
//========================================================================
{
	ULONG			tick_diff;
    LARGE_INTEGER	Time0, TimeRun;
	
	QueryPerformanceCounter(&Time0);
	do {
		QueryPerformanceCounter(&TimeRun);
		tick_diff   =  TimeRun.LowPart - Time0.LowPart;
	}while(tick_diff < tick_cnt);
}

//--------------------------------------------------------
// д
//--------------------------------------------------------
bool CI2C::i2cWriteData(int nLength, UCHAR *pData)
{
	int   i;
	LONG  err;
	CScopedLock autoLock(i2cLock());

	if(!autoLock.Status()) {
		_errNum = I2CERR_TIMEOUT;
		return false;
	}
	
	if(I2CSWStart()) {
 		for(i = 0; i < nLength;i++)	{
			if(!I2CSWWrite(*(pData+i)))	{
				break;
			}
		}
	}

	err = I2CGetLastError();
	I2CSWStop();
	//Sleep(10);
	return (err==I2CERR_OK);
}

//--------------------------------------------------------
// д
//--------------------------------------------------------
bool CI2C::i2cWriteDataNoStop(int nLength, UCHAR *pData)
{
	int   i;
	LONG  err;
	CScopedLock autoLock(i2cLock());
	
	if(!autoLock.Status()) {
		_errNum = I2CERR_TIMEOUT;
		return false;
	}
	
	if(I2CSWStart()) {
		for(i = 0; i < nLength;i++)	{
			if(!I2CSWWrite(*(pData+i)))	{
				break;
			}
		}
	}
	
	err = I2CGetLastError();

	return (err==I2CERR_OK);
}

//--------------------------------------------------------
// 
//--------------------------------------------------------
bool CI2C::i2cReadData(int nWriteBytes, UCHAR *cWrite, int nReadBytes, UCHAR *pData)
{	
	LONG  err;
	int   i;
	CScopedLock autoLock(i2cLock());
	
	if(!autoLock.Status()) {
		_errNum = I2CERR_TIMEOUT;
		return false;
	}
	
	if(I2CSWStart()) {
		for(i = 0; i < nWriteBytes;i++) {
			if(!I2CSWWrite(cWrite[i]))	{
				break;
			}
		}
	
		if(i >= nWriteBytes) {
			for(i = 0; i < nReadBytes; i++) {
				if(!I2CSWRead(pData+i)) {
					break;
				}
				if(i < nReadBytes - 1) {
					I2CSWSendACK();
				}
			}
			I2CSWSendNACK();
		}
	}

	err = I2CGetLastError();
	I2CSWStop();
	//Sleep(10);
	return (err==I2CERR_OK);
}

//--------------------------------------------------------
// 
//--------------------------------------------------------
bool CI2C::i2cReadData(UCHAR addr, UCHAR offset, int nReadBytes, UCHAR* pData)
{	
	LONG  err;
	int   i;
	UCHAR	Data[3];
	CScopedLock autoLock(i2cLock());
	
	if(!autoLock.Status()) {
		_errNum = I2CERR_TIMEOUT;
		return false;
	}
	
	Data[0] = addr;		//device addr
	Data[1] = offset;	//reg_addr
	
	if(I2CSWStart()) {
		for(i = 0; i < 2; i++) {
			if(!I2CSWWrite(Data[i]))	{
				break;
			}
		}
	}
	err = I2CGetLastError();
	I2CSWStop();

	if(err!=I2CERR_OK) {
		return false;
	}
		
	if(I2CSWStart()) {
		Data[0] |= 1;	//read addr
		if(I2CSWWrite(Data[0]))	{
			for(i = 0; i < nReadBytes; i++) {
				if(!I2CSWRead(pData+i)) {
					break;
				}
				if(i < nReadBytes - 1) {
					I2CSWSendACK();
				}
			}
			I2CSWSendNACK();
		}
	}
	
	err = I2CGetLastError();
	I2CSWStop();
	
	return (err==I2CERR_OK);
}

//--------------------------------------------------------
// Ժ
//--------------------------------------------------------
bool CI2C::i2cTest(UCHAR I2CAdd)
{	
	if(!I2CSWStart())		return false;
	if(!I2CSWWrite(I2CAdd)) return false;
	if(!I2CSWStop())		return false;

	return true;	
}

//--------------------------------------------------------
// I2Cд24C02Ժ
//--------------------------------------------------------
bool CI2C::i2cTest24C02()
{
	int i;
	BYTE data[5] = {0x60, 0x08, 0xB0, 0x01};
	
	//
	// д
	//

	if(I2CSWStart()) { 

		//д24C02ַ
		if(I2CSWWrite(0xA0)) { 

			//дsub addr
			if(I2CSWWrite(0xFC)) { 

				//д
				for(i=0; i<4; i++) {
					if(!I2CSWWrite(data[i])) { break; }
				}
			}
		}
	}

	if(!I2CSWStop()) { return false; }
	Sleep(40);

	//
	// 
	//

	if(I2CSWStart()) {

		//д24C02ַ
		if(I2CSWWrite(0xA0)) {

			//дsub addr
			if(I2CSWWrite(0xFC)) {

				if(I2CSWStart()) {
					//дaddr
					if(I2CSWWrite(0xA1)) {

						//
						for(i=0; i<4; i++) {
							if(!I2CSWRead(data+i)) { 
								break;
							}
							else {
								I2CSWSendACK();
							}
						}
						I2CSWSendNACK();
					}
				}
			}
		}
	}

	if(!I2CSWStop()) { return false; }

	return true;
}