//////////////////////////////////////////////////////////////////////
// I2C680X.cpp: implementation of the CI2C680X class.
// TW6801 I2Cд
// created by liran, 2007.01.19
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "I2C680X.h"
#include "TW680X.H"

//===========================================================================
// Tw68 I2C Class Implementation
//===========================================================================

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C680X::I2CInitHWMode( long freq )
//  Purpose: Initialize I2C for hardware control of SCL and SDA
//  Input:   long freq - frequency (hz) to run SCL at
//  Output:  None
//  Return:  None
/////////////////////////////////////////////////////////////////////////////
bool CI2C680X::I2CInitHWMode( long freq )
{
	//reset i2c master block
	m_pMem->outd(R_I2CRST, R_I2CRST_RST);
	
	// obtain data on data struct
	i2cShadow_SBUSD.addr_rw = 0;
	i2cShadow_SBUSD.byte1   = 0;
	i2cShadow_SBUSD.byte2   = 0;
	i2cShadow_SBUSD.rdb0	= 0;

	//set first byte
	m_pMem->outd(R_SBUSSD, *(ULONG*)&i2cShadow_SBUSD);

	// trigger control struct
	i2cShadow_SBUSC.value		   = 0;
	i2cShadow_SBUSC.SBUSC.sbmode   = 1;
	i2cShadow_SBUSC.SBUSC.wren	   = 0;
	i2cShadow_SBUSC.SBUSC.noack    = 1;
	i2cShadow_SBUSC.SBUSC.w3b      = 0;
	i2cShadow_SBUSC.SBUSC.bursten  = 0;
	i2cShadow_SBUSC.SBUSC.sync     = 0;
	i2cShadow_SBUSC.SBUSC.scl      = 1;
	i2cShadow_SBUSC.SBUSC.sda      = 1;
	i2cShadow_SBUSC.SBUSC.sbclk    = (freq==400000) ? 0x2A : 0xA6;
	i2cShadow_SBUSC.SBUSC.wdlen	   = 1;
	i2cShadow_SBUSC.SBUSC.rdlen	   = 1;
	i2cShadow_SBUSC.SBUSC.sbrw	   = 1;

	m_pMem->outd(R_SBUSC, i2cShadow_SBUSC.value);
	m_pMem->outd(R_SBUSTRIG, R_SBUSTRIG_SBTRIG);

	return CI2C::I2CInitHWMode(freq);
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C680X::I2CHWRead( BYTE address, BYTE *value )
//  Purpose: Perform a hardware read from the I2C
//  Input:   int address - address to be read from
//  Output:  int *value  - retrieved value
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C680X::I2CHWRead( BYTE address, BYTE *value )
{
	// check if correct mode is selected
	if ( _i2c_mode != I2CMode_HW ) {_errNum = I2CERR_MODE; return false;}

	I2C_SBUSD i2cTempD = i2cShadow_SBUSD;   
	// obtain data on data struct
	i2cTempD.addr_rw = address | I2C_READ;
	i2cTempD.byte1   = 0;
	i2cTempD.byte2   = 0;
	//set first byte
	m_pMem->outd(R_SBUSSD, *(ULONG *)&i2cTempD);

	I2C_SBUSC i2cTempC = i2cShadow_SBUSC.SBUSC;   
	// trigger control struct
	i2cTempC.w3b     = 0;
	i2cTempC.sbrw	 = 1;
	i2cTempC.rdlen	 = 1;
	i2cTempC.wren	 = 0;

	m_pMem->outd(R_SBUSC, *(ULONG *)&i2cTempC);

	if ( !I2CHWWaitUntilDone( TIMEOUT ) ) {_errNum = I2CERR_TIMEOUT; return false;}
	*value = (BYTE)(m_pMem->ind(R_SBUSSD) & R_SBUSRD_RB1);
	if ( !I2CHWReceivedACK() ) {_errNum = I2CERR_NOACK;	return false;}

	return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C680X::I2CHWWrite2( BYTE address, BYTE value1 )
//  Purpose:  Perform a hardware write of two bytes to the I2C
//  Input:   int address - address to be written to
//           int value1  - value of 2nd byte to be written
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C680X::I2CHWWrite2( BYTE address, BYTE value1 )
{
	if ( _i2c_mode != I2CMode_HW ) {	_errNum = I2CERR_MODE; return false;	}

	I2C_SBUSD i2cTempD = i2cShadow_SBUSD;   
	// obtain previous settings that didn't change

	i2cTempD.addr_rw = address | I2C_WRITE;
	i2cTempD.byte1   = value1;
	i2cTempD.byte2   = 0;
	//set i2c data register
	m_pMem->outd(R_SBUSSD, *(ULONG *)&i2cTempD);


	I2C_SBUSC i2cTempC = i2cShadow_SBUSC.SBUSC;   
	// obtain previous settings that didn't change
	i2cTempC.w3b     = 0;
	i2cTempC.wdlen	 = 1;
	i2cTempC.sbrw	 = 0;

	//set i2c control(trigger) register
	m_pMem->outd(R_SBUSC, *(ULONG *)&i2cTempC);

	if ( !I2CHWWaitUntilDone( TIMEOUT )) {	_errNum = I2CERR_TIMEOUT; return false; }
	if ( !I2CHWReceivedACK() == true )	 {	_errNum = I2CERR_NOACK;	return false; }
	
	return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C680X::I2CHWWrite3( BYTE address, BYTE value1, BYTE value2 )
//  Purpose: Perform a hardware write of three bytes to the I2C
//  Input:   int address - address to be written to
//           int value1  - value of 2nd byte to be written
//           int value2  - value of 3rd byte to be written
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C680X::I2CHWWrite3( BYTE address, BYTE value1, BYTE value2 )
{
	// check if correct mode is selected
	if ( _i2c_mode != I2CMode_HW ) {	_errNum = I2CERR_MODE; return false; }


	I2C_SBUSD i2cTempD = i2cShadow_SBUSD;   
	// obtain previous settings that didn't change

	i2cTempD.addr_rw = address | I2C_WRITE;
	i2cTempD.byte1   = value1;
	i2cTempD.byte2   = value2;

	m_pMem->outd(R_SBUSSD, *(ULONG *)&i2cTempD);


	I2C_SBUSC i2cTempC = i2cShadow_SBUSC.SBUSC;
	// obtain previous settings that didn't change
	i2cTempC.sbrw	 = 0;
	i2cTempC.w3b     = 1;
	i2cTempC.wdlen	 = 2;

	//set i2c control(trigger) register
	m_pMem->outd(R_SBUSC, *(ULONG *)&i2cTempC);


	if ( !I2CHWWaitUntilDone( TIMEOUT ) ) { _errNum = I2CERR_TIMEOUT; return false; }
	if ( !I2CHWReceivedACK() ) { _errNum = I2CERR_NOACK; return false; }

	return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  int CI2C680X::I2CReadSync( void )
//  Purpose: Read I2C sync value
//  Input:   None
//  Output:  None
//  Return:  Sync value
/////////////////////////////////////////////////////////////////////////////
int CI2C680X::I2CReadSync( void )
{
	return ((m_pMem->ind(R_SBUSC) & R_SBUSC_SYNC) >> 5);
}

//============================================================================
// Following functions are used internally
//============================================================================
   
/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C680X::I2CHWWaitUntilDone( int maxWait )
//  Purpose: Wait for hardware I2C to finish
//  Input:   int maxWait - maximum waiting time in milliseconds
//  Output:  None
//  Return:  true if done; false if timeout
/////////////////////////////////////////////////////////////////////////////
bool CI2C680X::I2CHWWaitUntilDone( int maxWait )
{
	// loop until either I2CDONE is set or timeout
	while (1)
	{
		if ( I2CHWIsDone() == true )
			return true;
#if 0
		// timeout?
		if ( TickCount() - startTime > (DWORD)maxWait )	{ _errNum = I2CERR_TIMEOUT; return false; }
#endif
	}
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C680X::I2CHWIsDone( void )
//  Purpose: Determine if I2C has finished a read or write operation by
//           checking the I2CDONE bit in the interrupt status register
//  Input:   None
//  Output:  None
//  Return:  true if done; else false
/////////////////////////////////////////////////////////////////////////////
bool CI2C680X::I2CHWIsDone( void )
{
	ULONG intStatus = m_pMem->ind(R_INTSTAT);
	if (intStatus & R_INT_SBDONE) {
		// Need to clear the bit when it is set; don't want to alter any other bits
		// Writing a 1 to the bit clears it.

		m_pMem->outd(R_INTSTAT, R_INT_SBDONE);
		return true;
	}
	return false;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C680X::I2CHWReceivedACK( void )
//  Purpose: Determine if ACK is receieved
//  Input:   None
//  Output:  None
//  Return:  true if ACK received; else false
/////////////////////////////////////////////////////////////////////////////
bool CI2C680X::I2CHWReceivedACK( void )
{
	ULONG intStatus = m_pMem->ind(R_INTSTAT);
	if(intStatus & R_INT_SBERR) {
		m_pMem->outd(R_INTSTAT, R_INT_SBERR);
		return false;
	}
	return true;
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C680X::I2CInitSWMode( long freq )
//  Purpose: Initialize I2C for software control of SCL and SDA
//  Input:   long freq - frequency (hz) to run SCL at
//  Output:  None
//  Return:  None
//  Note:    After calling I2CIsInitOK(), application should call this
//           function and check for return value is true before starting
//           software communication.
/////////////////////////////////////////////////////////////////////////////
bool CI2C680X::I2CInitSWMode( long freq )
{
	//reset i2c master block
	m_pMem->outd(R_I2CRST, R_I2CRST_RST);

	// obtain data on data struct
	i2cShadow_SBUSD.addr_rw = 0;
	i2cShadow_SBUSD.byte1   = 0;
	i2cShadow_SBUSD.byte2   = 0;
	i2cShadow_SBUSD.rdb0	= 0;

	//set first byte
	m_pMem->outd(R_SBUSSD, *(ULONG*)&i2cShadow_SBUSD);

	// trigger control struct
	i2cShadow_SBUSC.value		   = 0;
	i2cShadow_SBUSC.SBUSC.sbmode   = 0;
	i2cShadow_SBUSC.SBUSC.wren     = 0;
	i2cShadow_SBUSC.SBUSC.noack    = 1;
	i2cShadow_SBUSC.SBUSC.w3b      = 0;
	i2cShadow_SBUSC.SBUSC.bursten  = 0;
	i2cShadow_SBUSC.SBUSC.sync     = 0;
	i2cShadow_SBUSC.SBUSC.scl      = 1;
	i2cShadow_SBUSC.SBUSC.sda      = 1;
	i2cShadow_SBUSC.SBUSC.sbclk    = (freq==400000) ? 0x2A : 0xA6;
	i2cShadow_SBUSC.SBUSC.wdlen	   = 1;
	i2cShadow_SBUSC.SBUSC.rdlen	   = 1;
	i2cShadow_SBUSC.SBUSC.sbrw	   = 1;

	m_pMem->outd(R_SBUSC, i2cShadow_SBUSC.value);

	return CI2C::I2CInitSWMode(freq);
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C680X::I2CSWSetSCL( Level scl )
//  Purpose: Set software SCL value
//  Input:   Level scl - Hi releases the SCL output; Low forces the SCL output low
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C680X::I2CSWSetSCL( TLevel scl )
{
	i2cShadow_SBUSC.SBUSC.scl = scl;

	m_pMem->outd(R_SBUSC, i2cShadow_SBUSC.value);

	return CI2C::I2CSWSetSCL(scl);
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  int CI2C680X::I2CSWReadSCL( void )
//  Purpose: Read software SCL value
//  Input:   None
//  Output:  None
//  Return:  SCL value
/////////////////////////////////////////////////////////////////////////////
int CI2C680X::I2CSWReadSCL( void )
{
	return ((m_pMem->ind(R_SBUSC)&R_SBUSC_SSCLK) >> 6);
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  bool CI2C680X::I2CSWSetSDA( Level sda )
//  Purpose: Set software SDA value
//  Input:   Level sda - Hi releases the SDA output; Low forces the SDA output low
//  Output:  None
//  Return:  true or false
/////////////////////////////////////////////////////////////////////////////
bool CI2C680X::I2CSWSetSDA( TLevel sda )
{
	i2cShadow_SBUSC.SBUSC.sda = sda;

	m_pMem->outd(R_SBUSC, i2cShadow_SBUSC.value);

	return CI2C::I2CSWSetSDA(sda);
}

/////////////////////////////////////////////////////////////////////////////
//  Method:  int CI2C680X::I2CSWReadSDA( void )
//  Purpose: Read software SDA value
//  Input:   None
//  Output:  None
//  Return:  SDA value
/////////////////////////////////////////////////////////////////////////////
int CI2C680X::I2CSWReadSDA( void )
{
	return ((m_pMem->ind(R_SBUSC)&R_SBUSC_SSDAT) >> 7);
}